package com.inteligr8.alfresco.annotations.aspect;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public abstract class AbstractMethodAspect<A extends Annotation> extends AbstractWarnOnceService {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	private Set<Pair<String, String>> warned = new HashSet<>();
	
	protected A getAnnotation(ProceedingJoinPoint joinPoint, Class<A> annotationClass, boolean warnReturn, boolean warnThrows) {
		Method method = this.getMethod(joinPoint, annotationClass, warnReturn, warnThrows);
		return method.getAnnotation(annotationClass);
	}
	
	protected Method getMethod(ProceedingJoinPoint joinPoint, Object messagePrefix, boolean warnReturn, boolean warnThrows) {
		if (!(joinPoint.getSignature() instanceof MethodSignature))
			throw new IllegalStateException(this.createMessagePrefix(messagePrefix) + " must be on methods");
		
		MethodSignature methodSig = (MethodSignature) joinPoint.getSignature();
		if (warnReturn && methodSig.getReturnType() != null && !this.isReturnTypeVoid(methodSig)) {
			this.warn(joinPoint.toLongString(),
					"{} has a return value or throws clause; 'null' is always returned and subthread exceptions don't propagate: {}: {}",
					this.createMessagePrefix(messagePrefix), joinPoint, methodSig.getReturnType());
		}
		
		if (warnThrows && methodSig.getExceptionTypes() != null && methodSig.getExceptionTypes().length > 0) {
			this.warn(joinPoint.toLongString(),
					"{} has a return value or throws clause; 'null' is always returned and subthread exceptions don't propagate: {}: {}",
					this.createMessagePrefix(messagePrefix), joinPoint, methodSig.getExceptionTypes());
		}
		
		return methodSig.getMethod();
	}
	
	protected boolean isReturnTypeVoid(MethodSignature methodSig) {
		return Void.class.equals(methodSig.getReturnType()) || void.class.equals(methodSig.getReturnType());
	}
	
	private String createMessagePrefix(Object obj) {
		if (obj instanceof Class<?>) {
			Class<?> clazz = (Class<?>) obj;
			if (clazz.isAnnotation()) {
				return "The @" + ((Class<?>) obj).getSimpleName() + " annotated method";
			} else {
				return "The " + ((Class<?>) obj).getSimpleName() + " class method";
			}
		} else {
			return obj.toString();
		}
	}
	
	protected boolean validate(ProceedingJoinPoint joinPoint, Class<A> annotationClass,
			Collection<Class<? extends Annotation>> ignoredAnnotationClasses, Collection<Class<? extends Annotation>> disallowedAnnotationClasses) {
		Method method = this.getMethod(joinPoint, annotationClass, false, false);
		
		for (Class<? extends Annotation> a : ignoredAnnotationClasses) {
			if (method.isAnnotationPresent(a)) {
				if (this.warned.add(Pair.of(a.getName(), annotationClass.getName())))
					this.logger.warn("@{} cannot be used on the same method as @{}; ignoring annotation", a.getSimpleName(), annotationClass.getSimpleName());
			}
		}
		
		for (Class<? extends Annotation> a : disallowedAnnotationClasses) {
			if (method.isAnnotationPresent(a)) {
				this.logger.error("@{} cannot be used on the same method as @{}", a.getSimpleName(), annotationClass.getSimpleName());
				return false;
			}
		}
		
		return true;
	}

}
