package com.inteligr8.alfresco.annotations.aspect;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;

import org.alfresco.repo.transaction.AlfrescoTransactionSupport;
import org.alfresco.repo.transaction.AlfrescoTransactionSupport.TxnReadState;
import org.alfresco.repo.transaction.RetryingTransactionHelper;
import org.alfresco.repo.transaction.RetryingTransactionHelper.RetryingTransactionCallback;
import org.alfresco.service.transaction.TransactionService;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.transaction.IllegalTransactionStateException;
import org.springframework.transaction.annotation.Transactional;

import com.inteligr8.alfresco.annotations.TransactionalRetryable;
import com.inteligr8.alfresco.annotations.util.JtaTransactionalAnnotationAdapter;
import com.inteligr8.alfresco.annotations.util.SpringTransactionalAnnotationAdapter;
import com.inteligr8.alfresco.annotations.util.TransactionalAnnotationAdapter;

/**
 * This aspect implements the @Transactional and @TransactionalRetryable
 * annotations.
 * 
 * Most notably, it implements the Spring @Transactional annotation, so it
 * works when used within ACS modules.  Both could be used; or just either one.
 * Each situation has a different meaning.
 * 
 * - Without @TransactionalRetryable, it will not automatically retry due to
 * expected concurrency issues.
 * - Without @Transactional, it will be like a readonly
 * @Transactional(SUPPORTS)
 * 
 * @see org.springframework.transaction.annotation.Transactional
 * @see com.inteligr8.alfresco.annotations.TransactionalRetryable
 */
public abstract class AbstractRetryingTransactionAspect {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	private ApplicationContext context;
	
	@Autowired
	private TransactionService txService;
	
	public abstract String getJtaInterfaceName();
	
	/**
	 * A @Pointcunt annotation is not recognized in super classes.
	 */
	//@Pointcut("@annotation(org.springframework.transaction.annotation.Transactional) && execution(* *(..))")
	public abstract void isTransactionalAnnotated();
	
	public abstract void isJtaTransactionalAnnotated();

	/**
	 * A @Pointcunt annotation is not recognized in super classes.
	 */
	//@Pointcut("@annotation(com.inteligr8.alfresco.annotations.TransactionalRetryable) && execution(* *(..))")
	public abstract void isTransactionalRetryableAnnotated();

	/**
	 * An @Around annotation is not recognized in super classes.
	 */
	//@Around("isTransactionalAnnotated() || isJtaTransactionalAnnotated() || isTransactionalRetryableAnnotated()")
	public Object retryingTransactional(ProceedingJoinPoint joinPoint) throws Throwable {
		this.logger.trace("retryingTransactional({})", joinPoint);
		
		Method method = this.getMethod(joinPoint);
		TransactionalAnnotationAdapter txl = this.wrapTransactionalAnnotation(method);
		TransactionalRetryable txtry = method.getAnnotation(TransactionalRetryable.class);
		
		if (this.doCreateNewTxContext(txl) || this.isReadStateChange(txl)) {
			this.logger.debug("Changing TX context: {} => [ro: {}, new: {}]", AlfrescoTransactionSupport.getTransactionReadState(), txl.isReadOnly(), txl.getPropagation());
			return this.execute(joinPoint, txl, txtry);
		} else if (this.doCreateNewTxRetryContext(txtry)) {
			this.logger.debug("Changing TX context: retries: {}", txtry.maxRetries());
			return this.execute(joinPoint, null, txtry);
		} else {
			return joinPoint.proceed();
		}
	}
	
	private TransactionalAnnotationAdapter wrapTransactionalAnnotation(Method method) {
		Annotation txl = method.getAnnotation(Transactional.class);
		if (txl != null)
			return new SpringTransactionalAnnotationAdapter((Transactional) txl);
		
		txl = this.getOptionalAnnotation(method, this.getJtaInterfaceName());
		if (txl != null)
			return this.context.getAutowireCapableBeanFactory().getBean(JtaTransactionalAnnotationAdapter.class, txl);
		
		return null;
	}
	
	private <A extends Annotation> A getOptionalAnnotation(Method method, String fullyQualifiedAnnotationName) {
		try {
			@SuppressWarnings("unchecked")
			Class<A> annotationClass = (Class<A>) Class.forName(fullyQualifiedAnnotationName);
			return method.getAnnotation(annotationClass);
		} catch (ClassNotFoundException cnfe) {
			this.logger.trace("The {} annotation is not available in the classpath; assuming not set", fullyQualifiedAnnotationName);
			return null;
		}
	}
	
	private Method getMethod(ProceedingJoinPoint joinPoint) {
		if (!(joinPoint.getSignature() instanceof MethodSignature))
			throw new IllegalStateException("The @Transactional or @TransactionalRetryable annotations must be on methods");
		
		MethodSignature methodSig = (MethodSignature) joinPoint.getSignature();
		return methodSig.getMethod();
	}
	
	private boolean isReadStateChange(TransactionalAnnotationAdapter txl) {
		if (txl == null)
			return false;
		
		switch (txl.getPropagation()) {
			case NEVER:
			case NOT_SUPPORTED:
			case SUPPORTS:
				// do not force because of a read-state change
				return false;
			default:
		}
		
		switch (AlfrescoTransactionSupport.getTransactionReadState()) {
			case TXN_NONE:
				return true;
			case TXN_READ_ONLY:
				return !txl.isReadOnly();
			case TXN_READ_WRITE:
				return txl.isReadOnly();
			default:
				throw new IllegalStateException();
		}
	}
	
	private boolean doCreateNewTxRetryContext(TransactionalRetryable txtry) {
		return txtry != null;
	}
	
	private boolean doCreateNewTxContext(TransactionalAnnotationAdapter txl) {
		if (txl == null) {
			return false;
		} else switch (txl.getPropagation()) {
			case NEVER:
				switch (AlfrescoTransactionSupport.getTransactionReadState()) {
					case TXN_NONE:
						return false;
					default:
						throw new IllegalTransactionStateException("A transaction exists where one is not allowed");
				}
			case MANDATORY:
				switch (AlfrescoTransactionSupport.getTransactionReadState()) {
					case TXN_NONE:
						throw new IllegalTransactionStateException("A transaction does not exist where one is mandatory");
					case TXN_READ_ONLY:
						if (!txl.isReadOnly())
							throw new IllegalTransactionStateException("A read-only transaction exists where a read/write one is mandatory");
					case TXN_READ_WRITE:
						if (txl.isReadOnly())
							throw new IllegalTransactionStateException("A read/write transaction exists where a read-only one is mandatory");
				}
				return false;
			case NOT_SUPPORTED:
				switch (AlfrescoTransactionSupport.getTransactionReadState()) {
					case TXN_NONE:
						return false;
					default:
						throw new IllegalTransactionStateException("A transaction exists and pausing it is not supported");
				}
			case SUPPORTS:
				return false;
			case REQUIRED:
				switch (AlfrescoTransactionSupport.getTransactionReadState()) {
					case TXN_NONE:
						return true;
					default:
						return false;
				}
			case REQUIRES_NEW:
				return true;
			default:
				throw new IllegalTransactionStateException("The transactional propagation is not supported: " + txl.getPropagation());
		}
	}
	
	private Object execute(final ProceedingJoinPoint joinPoint, TransactionalAnnotationAdapter txl, TransactionalRetryable txtry) throws Throwable {
		RetryingTransactionCallback<Object> rtcallback = new RetryingTransactionCallback<Object>() {
			@Override
			public Object execute() throws Throwable {
				logger.debug("entering tx: {}", AlfrescoTransactionSupport.getTransactionId());
				
				try {
					return joinPoint.proceed();
				} catch (Exception | Error e) {
					throw e;
				} catch (Throwable t) {
					throw new RuntimeException("This should never happen", t);
				} finally {
					logger.trace("leaving tx: {}", AlfrescoTransactionSupport.getTransactionId());
				}
			}
		};
		
		RetryingTransactionHelper rthelper = new RetryingTransactionHelper();
		rthelper.setTransactionService(this.txService);
		if (txtry != null) {
			if (txtry.maxRetries() > 0)
				rthelper.setMaxRetries(txtry.maxRetries());
			if (txtry.minRetryWaitInMillis() > 0)
				rthelper.setMinRetryWaitMs(txtry.minRetryWaitInMillis());
			if (txtry.maxRetryWaitInMillis() > 0)
				rthelper.setMaxRetryWaitMs(txtry.maxRetryWaitInMillis());
			if (txtry.incRetryWaitInMillis() > 0)
				rthelper.setRetryWaitIncrementMs(txtry.incRetryWaitInMillis());
		}
		if (txl != null && txl.getTimeoutInSeconds() > 0)
			rthelper.setMaxExecutionMs(txl.getTimeoutInSeconds() * 1000L);
		
		try {
			this.logger.trace("source tx: {}", AlfrescoTransactionSupport.getTransactionId());
			boolean readonly = txl != null && txl.isReadOnly() || txl == null && AlfrescoTransactionSupport.getTransactionReadState() == TxnReadState.TXN_READ_ONLY;
			return rthelper.doInTransaction(rtcallback, readonly, txl != null);
		} catch (RuntimeException re) {
			// attempt to unwrap the exception
			if (re.getMessage() == null) {
				throw re;
			} else if (re.getMessage().startsWith("Exception from transactional callback")) {
				throw re.getCause();
			} else if (re.getMessage().startsWith("Exception in Transaction")) {
				throw re.getCause();
			} else {
				throw re;
			}
		} finally {
			this.logger.debug("returned to tx: {}", AlfrescoTransactionSupport.getTransactionId());
		}
	}

}
