package com.inteligr8.alfresco.annotations.aspect;

import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.DeclarePrecedence;
import org.aspectj.lang.annotation.Pointcut;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;

import com.inteligr8.alfresco.annotations.Asynchronous;
import com.inteligr8.alfresco.annotations.service.AsyncService;

/**
 * This aspect implements the @Asynchronous annotation.
 * 
 * @see com.inteligr8.alfresco.annotations.Asynchronous
 */
@Aspect
@DeclarePrecedence("com.inteligr8.alfresco.annotations.aspect.AsyncAspect, *")
public class AsyncAspect extends AbstractMethodAspect<Asynchronous> {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	@Qualifier("async.mq")
	private AsyncService durableAsyncService;

	@Autowired
	@Qualifier("async.thread")
	private AsyncService volatileAsyncService;
	
	@Pointcut("@annotation(com.inteligr8.alfresco.annotations.Asynchronous) && execution(* *(..))")
	public void isAsyncAnnotated() {
	}
	
	@Around("isAsyncAnnotated()")
	public Object async(ProceedingJoinPoint joinPoint) throws Throwable {
		this.logger.trace("async({})", joinPoint);
		
		Asynchronous async = this.getAnnotation(joinPoint, Asynchronous.class, true, true);
		
		AsyncService asyncService = async.durable() ? this.durableAsyncService : this.volatileAsyncService;
		if (!asyncService.isEnabled()) {
			this.warn(joinPoint.toLongString(),
					"Intercepted an @Asynchronous method call while the appropriate asynchronous service is not enabled; continuing synchronously");
			return joinPoint.proceed();
		} else if (asyncService.isCurrentThreadAsynchronous()) {
			this.logger.debug("Intercepted an @Asynchronous method call while already asynchronous; continuing synchronously");
			return joinPoint.proceed();
		} else {
			this.logger.trace("Intercepted an @Asynchronous method call; redirecting to the appropriate asynchronous service");
			asyncService.push(joinPoint);
			
			MethodSignature methodSig = (MethodSignature) joinPoint.getSignature();
			if (!this.isReturnTypeVoid(methodSig)) {
				this.warn(joinPoint.toLongString(),
						"An @Asynchronous method returns a value, which is not expected/allowed: {}",
						methodSig.getMethod());
			}
			return null;
		}
	}

}
