package com.inteligr8.alfresco.annotations.aspect;

import java.lang.reflect.Method;

import org.alfresco.repo.security.authentication.AuthenticationUtil;
import org.alfresco.repo.security.authentication.AuthenticationUtil.RunAsWork;
import org.apache.commons.lang3.StringUtils;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.DeclarePrecedence;
import org.aspectj.lang.annotation.Pointcut;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.inteligr8.alfresco.annotations.Authorizable;
import com.inteligr8.alfresco.annotations.Authorized;
import com.inteligr8.alfresco.annotations.AuthorizedAsSystem;

/**
 * This aspect implements the Authorized and AuthorizedAsSystem annotations.
 * 
 * @see com.inteligr8.alfresco.annotations.Authorized
 * @see com.inteligr8.alfresco.annotations.AuthorizedAsSystem
 */
@Aspect
@DeclarePrecedence("com.inteligr8.alfresco.annotations.aspect.AuthorizedAspect, com.inteligr8.alfresco.annotations.aspect.RetryingTransactionAspect")
public class AuthorizedAspect {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Pointcut("@annotation(com.inteligr8.alfresco.annotations.Authorized) && execution(* *(..))")
	public void isAuthorizedAnnotated() {
	}

	@Pointcut("@annotation(com.inteligr8.alfresco.annotations.AuthorizedAsSystem) && execution(* *(..))")
	public void isAuthorizedAsSystemAnnotated() {
	}

	@Around("isAuthorizedAnnotated() || isAuthorizedAsSystemAnnotated()")
	public Object runAs(ProceedingJoinPoint joinPoint) throws Throwable {
		this.logger.trace("runAs({})", joinPoint);
		
		String runAsUser = this.getRunAsUser(joinPoint);

		String currentRunAsUser = AuthenticationUtil.getRunAsUser();
		if (currentRunAsUser != null && currentRunAsUser.equals(runAsUser)) {
			this.logger.trace("The current context is already running as the specified user: {}", currentRunAsUser);
			return joinPoint.proceed();
		} else {
			this.logger.debug("Changing runAs context: {} => {}", currentRunAsUser, runAsUser);
			return this.runAs(joinPoint, runAsUser);
		}
	}
	
	private String getRunAsUser(ProceedingJoinPoint joinPoint) {
		if (!(joinPoint.getSignature() instanceof MethodSignature))
			throw new IllegalStateException("The @Authorized annotation must be on methods and methods have signatures");
		
		MethodSignature methodSig = (MethodSignature) joinPoint.getSignature();
		Method method = methodSig.getMethod();
		if (method.getAnnotation(AuthorizedAsSystem.class) != null) {
			String runAs = AuthenticationUtil.getSystemUserName();
			this.logger.trace("The @AuthorizedAsSystem method '{}' will run as: {}", method, runAs);
			return runAs;
		}
		
		if (joinPoint.getThis() instanceof Authorizable) {
			String runAs = StringUtils.trimToNull(((Authorizable) joinPoint.getThis()).authorizeAsUser());
			if (runAs == null)
				throw new IllegalArgumentException("A 'null' value is allowed from authorizeAsUser()");
			
			this.logger.trace("The @Authorized method '{}' is Authorizable: {}", method, runAs);
			return runAs;
		}
		
		Authorized runAsAnnotation = method.getAnnotation(Authorized.class);
		String runAs = StringUtils.trimToNull(runAsAnnotation.value());
		if (runAs != null && runAs.length() > 0) {
			this.logger.trace("The @Authorized method '{}' must run as: {}", method, runAs);
			return runAs;
		}

		this.logger.trace("The @Authorized method '{}' must run as system", method);
		return AuthenticationUtil.getSystemUserName();
	}
	
	private Object runAs(final ProceedingJoinPoint joinPoint, String runAsUser) throws Throwable {
		RunAsWork<Object> work = new RunAsWork<Object>() {
			public Object doWork() throws Exception {
				try {
					return joinPoint.proceed();
				} catch (Exception | Error e) {
					throw e;
				} catch (Throwable t) {
					throw new RuntimeException("This should never happen", t);
				}
			}
		};
		
		try {
			return AuthenticationUtil.runAs(work, runAsUser);
		} catch (RuntimeException re) {
			// attempt to unwrap the exception
			if (re.getMessage() != null && re.getMessage().equals("Error during run as.")) {
				throw re.getCause();
			} else {
				throw re;
			}
		}
	}

}
