package com.inteligr8.alfresco.annotations.aspect;

import java.lang.reflect.Method;

import org.alfresco.repo.lock.JobLockService;
import org.alfresco.service.namespace.QName;
import org.apache.commons.lang3.StringUtils;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.DeclarePrecedence;
import org.aspectj.lang.annotation.Pointcut;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import com.inteligr8.alfresco.annotations.JobSynchronized;

/**
 * This aspect implements the JobSynchronized annotation.
 * 
 * @see com.inteligr8.alfresco.annotations.JobSynchronized
 */
@Aspect
@DeclarePrecedence("com.inteligr8.alfresco.annotations.aspect.RetryingTransactionAspect, com.inteligr8.alfresco.annotations.aspect.JobLockAspect")
public class JobLockAspect extends AbstractMethodAspect<JobSynchronized> {

	private static final String NS = "http://inteligr8.com/alfresco/model";
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	private JobLockService jobLockService;
	
	@Pointcut("@annotation(com.inteligr8.alfresco.annotations.JobSynchronized) && execution(* *(..))")
	public void isJobSyncAnnotated() {
	}
	
	@Around("isJobSyncAnnotated()")
	public Object jobSync(ProceedingJoinPoint joinPoint) throws Throwable {
		this.logger.trace("jobSync({})", joinPoint);

		Method method = this.getMethod(joinPoint, JobSynchronized.class, false, false);
		JobSynchronized clusterSync = method.getAnnotation(JobSynchronized.class);
		
		QName lockQName = this.getLockQName(clusterSync, method);
		
		this.logger.debug("Acquiring job lock: {}", lockQName);
		String lockToken = this.jobLockService.getLock(lockQName, clusterSync.lockTimeoutInMillis(),
				clusterSync.acquireWaitBetweenRetriesInMillis(), clusterSync.acquireMaxRetries());
		try {
			this.logger.trace("Acquired job lock: {}", lockQName);
			
			return joinPoint.proceed();
		} finally {
			this.logger.debug("Releasing job lock: {}", lockQName);
			this.jobLockService.releaseLock(lockToken, lockQName);
		}
	}
	
	private QName getLockQName(JobSynchronized clusterSync, Method method) {
		String lockName = StringUtils.trimToNull(clusterSync.value());
		if (lockName != null) {
			return QName.createQNameWithValidLocalName(NS, lockName);
		} else {
			String methodId = method.getDeclaringClass().getSimpleName() + "_" + method.getName();
			return QName.createQNameWithValidLocalName(NS, methodId);
		}
	}

}
