package com.inteligr8.alfresco.annotations.aspect;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Collections;
import java.util.Set;

import org.alfresco.service.cmr.dictionary.ClassDefinition;
import org.alfresco.service.cmr.dictionary.DictionaryService;
import org.alfresco.service.cmr.repository.NodeRef;
import org.alfresco.service.cmr.repository.NodeService;
import org.alfresco.service.namespace.QName;
import org.alfresco.service.namespace.QNamePattern;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.DeclarePrecedence;
import org.aspectj.lang.annotation.Pointcut;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import com.inteligr8.alfresco.annotations.IfNodeHasAspect;
import com.inteligr8.alfresco.annotations.NodeAspectConstrainable;

/**
 * This aspect implements the IfNodeHasAspect annotation.
 * 
 * @see com.inteligr8.alfresco.annotations.IfNodeHasAspect
 */
@Aspect
@DeclarePrecedence("com.inteligr8.alfresco.annotations.aspect.RetryingTransactionAspect, com.inteligr8.alfresco.annotations.aspect.NodeAspectAspect")
public class NodeAspectAspect extends QNameBasedAspect<IfNodeHasAspect> {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	private DictionaryService dictionaryService;
	
	@Autowired
	private NodeService nodeService;
	
	@Value("${inteligr8.cache.nodeAspectConstrainable.maxBeans}")
	private int maxBeans;
	
	@Override
	public int getMaxBeansCache() {
		return this.maxBeans;
	}
	
	@Override
	public Class<IfNodeHasAspect> getAnnotationClass() {
		return IfNodeHasAspect.class;
	}
	
	@Pointcut("@annotation(com.inteligr8.alfresco.annotations.IfNodeHasAspect) && execution(* *(..))")
	public void isIfAspectAnnotated() {
	}

	@Pointcut("execution(* *(@com.inteligr8.alfresco.annotations.IfNodeHasAspect (*), ..))")
	public void isIfAspectParamAnnotated() {
	}
	
	@Around("isIfAspectAnnotated() || isIfAspectParamAnnotated()")
	public Object isNodeAspect(ProceedingJoinPoint joinPoint) throws Throwable {
		this.logger.trace("isNodeAspect({})", joinPoint);
		
		return this.checkParameters(joinPoint, new ApplicableParameterCallback<IfNodeHasAspect>() {
			@Override
			public boolean checkParameter(ProceedingJoinPoint joinPoint, Method method, IfNodeHasAspect annotation, int parameterIndex) {
				Object arg = joinPoint.getArgs()[parameterIndex];
				Collection<NodeRef> nodeRefs = extractNodeRefs(arg);
				if (nodeRefs == null)
					return true;

				QNameBasedCallback<IfNodeHasAspect> callback = new QNameBasedCallback<IfNodeHasAspect>() {
					@Override
					public boolean isConstrained() {
						return joinPoint.getThis() instanceof NodeAspectConstrainable;
					}
					
					@Override
					public String getConstrainableClassSimpleName() {
						return NodeAspectConstrainable.class.getSimpleName();
					}
					
					@Override
					public String getBeanName() {
						return ((NodeAspectConstrainable) joinPoint.getThis()).getBeanName();
					}
					
					@Override
					public Collection<? extends QNamePattern> constrainedQNames() {
						return ((NodeAspectConstrainable) joinPoint.getThis()).constrainedAspects();
					}
					
					@Override
					public Collection<QName> allPossibleQNames() {
						return dictionaryService.getAllAspects();
					}
					
					@Override
					public void addAllAncestors(Set<QName> qnames, QName qname) {
						ClassDefinition aspectDef = dictionaryService.getAspect(qname);
						while (aspectDef != null) {
							qnames.add(aspectDef.getName());
							aspectDef = aspectDef.getParentClassDefinition();
						}
					}
					
					@Override
					public String getAnnotationValue(IfNodeHasAspect annotation) {
						return annotation.aspect();
					}
				};
				
				Set<QName> aspects = getQNameCache(joinPoint, annotation, callback);
				for (NodeRef nodeRef : nodeRefs) {
					Set<QName> nodeAspects = nodeService.getAspects(nodeRef);
					if (Collections.disjoint(aspects, nodeAspects)) {
						logger.debug("The node '{}' aspects {} are not applicable; skipping method: {}", nodeRef, aspects, method);
						return false;
					}
				}
				
				return true;
			}
		});
	}

}
