package com.inteligr8.alfresco.annotations.aspect;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Set;

import org.alfresco.service.cmr.dictionary.ClassDefinition;
import org.alfresco.service.cmr.dictionary.DictionaryService;
import org.alfresco.service.cmr.repository.NodeRef;
import org.alfresco.service.cmr.repository.NodeService;
import org.alfresco.service.namespace.QName;
import org.alfresco.service.namespace.QNamePattern;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.DeclarePrecedence;
import org.aspectj.lang.annotation.Pointcut;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import com.inteligr8.alfresco.annotations.IfNodeOfType;
import com.inteligr8.alfresco.annotations.NodeTypeConstrainable;

/**
 * This aspect implements the IfNodeOfType annotation.
 * 
 * @see com.inteligr8.alfresco.annotations.IfNodeOfType
 */
@Aspect
@DeclarePrecedence("com.inteligr8.alfresco.annotations.aspect.RetryingTransactionAspect, com.inteligr8.alfresco.annotations.aspect.NodeTypeAspect")
public class NodeTypeAspect extends QNameBasedAspect<IfNodeOfType> {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	private DictionaryService dictionaryService;
	
	@Autowired
	private NodeService nodeService;
	
	@Value("${inteligr8.cache.nodeTypeConstrainable.maxBeans}")
	private int maxBeans;
	
	@Override
	public int getMaxBeansCache() {
		return this.maxBeans;
	}
	
	@Override
	public Class<IfNodeOfType> getAnnotationClass() {
		return IfNodeOfType.class;
	}
	
	@Pointcut("@annotation(com.inteligr8.alfresco.annotations.IfNodeOfType) && execution(* *(..))")
	public void isIfNodeTypeAnnotated() {
	}

	@Pointcut("execution(* *(@com.inteligr8.alfresco.annotations.IfNodeOfType (*), ..))")
	public void isIfNodeTypeParamAnnotated() {
	}
	
	@Around("isIfNodeTypeAnnotated() || isIfNodeTypeParamAnnotated()")
	public Object isNodeType(ProceedingJoinPoint joinPoint) throws Throwable {
		this.logger.trace("isNodeType({})", joinPoint);
		
		return this.checkParameters(joinPoint, new ApplicableParameterCallback<IfNodeOfType>() {
			@Override
			public boolean checkParameter(ProceedingJoinPoint joinPoint, Method method, IfNodeOfType annotation, int parameterIndex) {
				Object arg = joinPoint.getArgs()[parameterIndex];
				Collection<NodeRef> nodeRefs = extractNodeRefs(arg);
				if (nodeRefs == null)
					return true;

				QNameBasedCallback<IfNodeOfType> callback = new QNameBasedCallback<IfNodeOfType>() {
					@Override
					public boolean isConstrained() {
						return joinPoint.getTarget() instanceof NodeTypeConstrainable;
					}
					
					@Override
					public String getConstrainableClassSimpleName() {
						return NodeTypeConstrainable.class.getSimpleName();
					}
					
					@Override
					public String getBeanName() {
						return ((NodeTypeConstrainable) joinPoint.getTarget()).getBeanName();
					}
					
					@Override
					public Collection<? extends QNamePattern> constrainedQNames() {
						return ((NodeTypeConstrainable) joinPoint.getTarget()).constrainedNodeTypes();
					}
					
					@Override
					public Collection<QName> allPossibleQNames() {
						return dictionaryService.getAllTypes();
					}
					
					@Override
					public void addAllAncestors(Set<QName> qnames, QName qname) {
						ClassDefinition typeDef = dictionaryService.getType(qname);
						while (typeDef != null) {
							qnames.add(typeDef.getName());
							typeDef = typeDef.getParentClassDefinition();
						}
					}
					
					@Override
					public String getAnnotationValue(IfNodeOfType annotation) {
						return annotation.type();
					}
				};
				
				Set<QName> nodeTypes = getQNameCache(joinPoint, annotation, callback);
				for (NodeRef nodeRef : nodeRefs) {
					QName nodeType = nodeService.getType(nodeRef);
					if (!nodeTypes.contains(nodeType)) {
						logger.debug("The node '{}' type '{}' is not applicable; skipping method: {}", nodeRef, nodeType, method);
						return false;
					}
				}
				
				return true;
			}
		});
	}

}
