package com.inteligr8.alfresco.annotations.aspect;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;

import org.alfresco.service.cmr.repository.AssociationRef;
import org.alfresco.service.cmr.repository.ChildAssociationRef;
import org.alfresco.service.cmr.repository.NodeRef;
import org.alfresco.service.cmr.repository.NodeService;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.DeclarePrecedence;
import org.aspectj.lang.annotation.Pointcut;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import com.inteligr8.alfresco.annotations.IfNodeExists;

/**
 * This aspect implements the IfNodeExists annotation.
 * 
 * @see com.inteligr8.alfresco.annotations.IfNodeExists
 */
@Aspect
@DeclarePrecedence("com.inteligr8.alfresco.annotations.aspect.RetryingTransactionAspect, com.inteligr8.alfresco.annotations.aspect.OperableNodeAspect")
public class OperableNodeAspect extends AbstractMethodOrParameterAspect<IfNodeExists> {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	private NodeService nodeService;
	
	@Override
	public Class<IfNodeExists> getAnnotationClass() {
		return IfNodeExists.class;
	}
	
	@Pointcut("@annotation(com.inteligr8.alfresco.annotations.IfNodeExists) && execution(* *(..))")
	public void isIfNodeExistsAnnotated() {
	}

	@Pointcut("execution(* *(@com.inteligr8.alfresco.annotations.IfNodeExists (*), ..))")
	public void isIfNodeExistsParamAnnotated() {
	}
	
	@Around("isIfNodeExistsAnnotated() || isIfNodeExistsParamAnnotated()")
	public Object isNodeOperable(ProceedingJoinPoint joinPoint) throws Throwable {
		return this.checkParameters(joinPoint, new ApplicableParameterCallback<IfNodeExists>() {
			@Override
			public boolean checkParameter(ProceedingJoinPoint joinPoint, Method method, IfNodeExists annotation, int parameterIndex) {
				Object arg = joinPoint.getArgs()[parameterIndex];
				Collection<NodeRef> nodeRefs = extractNodeRefs(arg);
				if (nodeRefs == null)
					return true;

				logger.trace("Checking if nodes are operable: {}", nodeRefs);
				
				for (NodeRef nodeRef : nodeRefs) {
					if (!isOperableNode(nodeRef)) {
						logger.debug("The node '{}' does not exist; skipping method: {}", nodeRef, method);
						return false;
					}
				}
				
				return true;
			}
		});
	}
	
	private Collection<NodeRef> extractNodeRefs(Object obj) {
		if (obj instanceof NodeRef) {
			NodeRef nodeRef = (NodeRef) obj;
			return Collections.singleton(nodeRef);
		} else if (obj instanceof ChildAssociationRef) {
			ChildAssociationRef childAssocRef = (ChildAssociationRef) obj;
			return Collections.singleton(childAssocRef.getChildRef());
		} else if (obj instanceof AssociationRef) {
			AssociationRef assocRef = (AssociationRef) obj;
			return Arrays.asList(assocRef.getSourceRef(), assocRef.getTargetRef());
		} else if (obj instanceof Collection<?>) {
			Set<NodeRef> nodeRefs = new LinkedHashSet<>();
			for (Object o : ((Collection<?>) obj)) {
				Collection<NodeRef> subNodeRefs = this.extractNodeRefs(o);
				if (subNodeRefs != null)
					nodeRefs.addAll(subNodeRefs);
			}
			return nodeRefs;
		} else if (obj instanceof Object[]) {
			Set<NodeRef> nodeRefs = new LinkedHashSet<>();
			for (Object o : ((Object[]) obj)) {
				Collection<NodeRef> subNodeRefs = this.extractNodeRefs(o);
				if (subNodeRefs != null)
					nodeRefs.addAll(subNodeRefs);
			}
			return nodeRefs;
		} else {
			return null;
		}
	}
	
	private boolean isOperableNode(NodeRef nodeRef) {
		if (!this.nodeService.exists(nodeRef)) {
			this.logger.debug("The node '{}' does not exist", nodeRef);
			return false;
		} else if (this.nodeService.getNodeStatus(nodeRef).isDeleted()) {
			this.logger.debug("The node '{}' was already deleted", nodeRef);
			return false;
		}
		
		return true;
	}

}
