package com.inteligr8.alfresco.annotations.aspect;

import java.lang.annotation.Annotation;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;

import org.alfresco.repo.cache.DefaultSimpleCache;
import org.alfresco.repo.cache.SimpleCache;
import org.alfresco.service.cmr.repository.AssociationRef;
import org.alfresco.service.cmr.repository.ChildAssociationRef;
import org.alfresco.service.cmr.repository.NodeRef;
import org.alfresco.service.namespace.NamespaceService;
import org.alfresco.service.namespace.QName;
import org.alfresco.service.namespace.QNamePattern;
import org.aspectj.lang.ProceedingJoinPoint;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import jakarta.annotation.PostConstruct;

public abstract class QNameBasedAspect<T extends Annotation> extends AbstractMethodOrParameterAspect<T> {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired
	private NamespaceService namespaceService;
	
	private SimpleCache<String, Set<QName>> qnameCache;
	
	public abstract int getMaxBeansCache();
	
	@PostConstruct
	public void init() {
		this.qnameCache = new DefaultSimpleCache<>(this.getMaxBeansCache(), this.getClass().getName());
	}
	
	public Set<QName> getQNameCache(ProceedingJoinPoint joinPoint, T annotation, QNameBasedCallback<T> callback) {
		if (callback.isConstrained()) {
			Set<QName> qnames = this.qnameCache.get(callback.getBeanName());
			if (qnames != null) {
				this.logger.trace("Using cache of qnames for bean: {}", callback.getBeanName());
				return qnames;
			}
			
			// caching all types; expensive now; fast at runtime
			qnames = new HashSet<>();
			for (QNamePattern qnamePattern : callback.constrainedQNames()) {
				if (qnamePattern instanceof QName) {
					callback.addAllAncestors(qnames, (QName) qnamePattern);
				} else {
					for (QName qname : callback.allPossibleQNames()) {
						if (qnamePattern.isMatch(qname)) {
							callback.addAllAncestors(qnames, qname);
						}
					}
				}
			}
			
			this.logger.debug("Caching @{} qnames for bean: {}: {}", annotation.getClass().getSimpleName(), callback.getBeanName(), qnames);
			this.qnameCache.put(callback.getBeanName(), qnames);
			return qnames;
		} else if (callback.getAnnotationValue(annotation).length() > 0) {
			Set<QName> qnames = this.qnameCache.get(joinPoint.getThis().getClass().getName());
			if (qnames != null) {
				this.logger.trace("Using cache of qnames for bean: {}", joinPoint.getThis().getClass());
				return qnames;
			}
			
			qnames = new HashSet<>();
			callback.addAllAncestors(qnames, QName.createQName(callback.getAnnotationValue(annotation), this.namespaceService));

			this.logger.debug("Caching @{} node types for singleton: {}: {}", annotation.getClass().getSimpleName(), joinPoint.getThis().getClass(), qnames);
			this.qnameCache.put(joinPoint.getThis().getClass().getName(), qnames);
			return qnames;
		} else {
			throw new IllegalStateException("An @" + annotation.getClass().getSimpleName() + " must have a value or the class must implement " + callback.getConstrainableClassSimpleName());
		}
	}
	
	protected Collection<NodeRef> extractNodeRefs(Object obj) {
		if (obj instanceof NodeRef) {
			NodeRef nodeRef = (NodeRef) obj;
			return Collections.singleton(nodeRef);
		} else if (obj instanceof ChildAssociationRef) {
			ChildAssociationRef childAssocRef = (ChildAssociationRef) obj;
			return Collections.singleton(childAssocRef.getChildRef());
		} else if (obj instanceof AssociationRef) {
			AssociationRef assocRef = (AssociationRef) obj;
			return Arrays.asList(assocRef.getSourceRef(), assocRef.getTargetRef());
		} else if (obj instanceof Collection<?>) {
			Set<NodeRef> nodeRefs = new LinkedHashSet<>();
			for (Object o : ((Collection<?>) obj)) {
				Collection<NodeRef> subNodeRefs = this.extractNodeRefs(o);
				if (subNodeRefs != null)
					nodeRefs.addAll(subNodeRefs);
			}
			return nodeRefs;
		} else if (obj instanceof Object[]) {
			Set<NodeRef> nodeRefs = new LinkedHashSet<>();
			for (Object o : ((Object[]) obj)) {
				Collection<NodeRef> subNodeRefs = this.extractNodeRefs(o);
				if (subNodeRefs != null)
					nodeRefs.addAll(subNodeRefs);
			}
			return nodeRefs;
		} else {
			return null;
		}
	}
	
	
	
	public interface QNameBasedCallback<T> {
		
		boolean isConstrained();
		
		String getBeanName();
		
		String getConstrainableClassSimpleName();
		
		Collection<? extends QNamePattern> constrainedQNames();
		
		void addAllAncestors(Set<QName> qnames, QName qname);
		
		Collection<QName> allPossibleQNames();
		
		String getAnnotationValue(T annotation);
		
	}

}
