package com.inteligr8.alfresco.annotations.aspect;

import java.lang.Thread.UncaughtExceptionHandler;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.Supplier;

import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.DeclarePrecedence;
import org.aspectj.lang.annotation.Pointcut;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.inteligr8.alfresco.annotations.Threadable;
import com.inteligr8.alfresco.annotations.Threaded;

/**
 * This aspect implements the @Threaded annotation.
 * 
 * @see com.inteligr8.alfresco.annotations.Threaded
 */
@Aspect
@DeclarePrecedence("com.inteligr8.alfresco.annotations.aspect.AsyncAspect, com.inteligr8.alfresco.annotations.aspect.ThreadedAspect, *")
public class ThreadedAspect extends AbstractMethodAspect<Threaded> {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	private ThreadLocal<Set<String>> nested = ThreadLocal.withInitial(new Supplier<Set<String>>() {
		public Set<String> get() {
			return new HashSet<>();
		}
	});
	
	@Pointcut("@annotation(com.inteligr8.alfresco.annotations.Threaded) && execution(* *(..))")
	public void isThreadedAnnotated() {
	}
	
	@Around("isThreadedAnnotated()")
	public Object threaded(ProceedingJoinPoint joinPoint) throws Throwable {
		// AspectJ does not recursively match annotations if the same thread is calling joinPoint.proceed()
		// but when a different thread calls it, it doesn't know it is already processing the annotation
		// so we are going to use a ThreadLocal to prevent recursion across threads
		if (this.nested.get().contains(joinPoint.getSignature().toLongString()))
			return joinPoint.proceed();
		
		this.logger.trace("threaded({})", joinPoint);
		
		Threaded threaded = this.getAnnotation(joinPoint, Threaded.class, true, true);
		
		MergedThreadConfiguration threadConfig = new MergedThreadConfiguration(joinPoint, threaded);
		
		ThreadFactoryBuilder tfbuilder = new ThreadFactoryBuilder()
				.setPriority(threadConfig.getThreadPriority())
				.setUncaughtExceptionHandler(new UncaughtExceptionHandler() {
					@Override
					public void uncaughtException(Thread t, Throwable e) {
						logger.error("The thread '" + t.getName() + "' had an exception", e);
					}
				});
		if (threaded.name().length() > 0)
			tfbuilder.setNameFormat(threaded.name() + "-%d");
		
		Integer concurrency = threadConfig.getConcurrency();
		if (concurrency == null)
			concurrency = threadConfig.getThreads();
		BlockingQueue<Runnable> threadQueue = new ArrayBlockingQueue<>(threadConfig.getThreads().intValue());
		
		ThreadPoolExecutor threadExecutor = new ThreadPoolExecutor(concurrency.intValue(), concurrency.intValue(),
				1L, TimeUnit.SECONDS,
				threadQueue,
				tfbuilder.build());

		Callable<Object> callable = this.createCallable(joinPoint);

		this.logger.debug("Starting {} threads", threadConfig.getThreads());
		for (int t = 0; t < threadConfig.getThreads().intValue(); t++)
			threadExecutor.submit(callable);

		threadExecutor.shutdown();
		
		if (threaded.join()) {
			long waitMillis = threaded.joinWaitMillis() == 0L ? 300000L : threaded.joinWaitMillis();
			while (true) {
				this.logger.debug("Blocking this thread until subthreads finish: {}", Thread.currentThread().getId());
				if (!threadExecutor.awaitTermination(waitMillis, TimeUnit.MILLISECONDS)) {
					if (threaded.joinWaitMillis() > 0L)
						throw new TimeoutException();
				} else {
					this.logger.trace("Subthreads finished; unblocking this thread: {}", Thread.currentThread().getId());
					break;
				}
			}

			this.logger.debug("Subthreads running: {}; unblocking this thread: {}", threadExecutor.getActiveCount(), Thread.currentThread().getId());
		}
		
		return null;
	}
	
	private Callable<Object> createCallable(final ProceedingJoinPoint joinPoint) throws Throwable {
		return new Callable<Object>() {
			@Override
			public Object call() throws Exception {
				logger.debug("entering thread: {}", Thread.currentThread().getId());
				
				// AspectJ does not recursively match annotations if the same thread is calling joinPoint.proceed()
				// but when a different thread calls it, it doesn't know it is already processing the annotation
				// so we are going to use a ThreadLocal to prevent recursion across threads
				nested.get().add(joinPoint.getSignature().toLongString());
				
				// we cannot use joinPoint.proceed() as it will only execute in 1 thread
				// we need to recreate the method call, fresh, using reflection
				MethodSignature methodSig = (MethodSignature) joinPoint.getSignature();
				Method method = methodSig.getMethod();
				
				try {
					method.setAccessible(true);
					return method.invoke(joinPoint.getThis(), joinPoint.getArgs());
				} catch (Exception | Error e) {
					logger.error("An unexpected issue occurred in thread #" + Thread.currentThread().getId(), e);
					throw e;
				} catch (Throwable t) {
					logger.error("An unexpected issue occurred in thread #" + Thread.currentThread().getId(), t);
					throw new RuntimeException("This should never happen", t);
				} finally {
					logger.trace("leaving thread: {}", Thread.currentThread().getId());
				}
			}
		};
	}
	
	
	
	private class MergedThreadConfiguration implements Threadable {
		
		private final Threaded threaded;
		private final Threadable threadable;
		
		public MergedThreadConfiguration(ProceedingJoinPoint joinPoint, Threaded threaded) {
			this.threaded = threaded;
			this.threadable = (joinPoint.getThis() instanceof Threadable) ? (Threadable) joinPoint.getThis() : null;
		}
		
		@Override
		public Integer getThreads() {
			if (this.threadable != null && this.threadable.getThreads() != null) {
				return this.threadable.getThreads();
			} else {
				return this.threaded.threads();
			}
		}
		
		@Override
		public Integer getConcurrency() {
			if (this.threadable != null && this.threadable.getConcurrency() != null) {
				return this.threadable.getConcurrency();
			} else if (this.threaded.concurrency() <= 0) {
				return null;
			} else {
				return this.threaded.concurrency();
			}
		}
		
		@Override
		public Integer getThreadPriority() {
			if (this.threadable != null && this.threadable.getThreadPriority() != null) {
				return this.threadable.getThreadPriority();
			} else {
				return this.threaded.priority();
			}
		}
		
	}

}
