package com.inteligr8.alfresco.annotations.service.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Parameter;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.OffsetDateTime;
import java.time.OffsetTime;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.Temporal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Supplier;
import java.util.regex.Pattern;

import org.alfresco.error.AlfrescoRuntimeException;
import org.alfresco.model.ContentModel;
import org.alfresco.repo.cache.SimpleCache;
import org.alfresco.repo.dictionary.M2Model;
import org.alfresco.repo.version.common.VersionImpl;
import org.alfresco.service.cmr.action.Action;
import org.alfresco.service.cmr.action.ActionService;
import org.alfresco.service.cmr.dictionary.CustomModelService;
import org.alfresco.service.cmr.dictionary.DictionaryService;
import org.alfresco.service.cmr.repository.ContentReader;
import org.alfresco.service.cmr.repository.ContentService;
import org.alfresco.service.cmr.repository.NodeRef;
import org.alfresco.service.cmr.version.Version;
import org.alfresco.service.namespace.NamespaceService;
import org.alfresco.service.namespace.QName;
import org.alfresco.service.transaction.TransactionService;
import org.alfresco.util.Pair;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.reflect.MethodSignature;
import org.quartz.JobKey;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.Trigger;
import org.quartz.TriggerBuilder;
import org.quartz.impl.JobDetailImpl;
import org.quartz.impl.StdSchedulerFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEvent;
import org.springframework.extensions.surf.util.AbstractLifecycleBean;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.inteligr8.alfresco.annotations.Threadable;
import com.inteligr8.alfresco.annotations.job.AsyncJob;
import com.inteligr8.alfresco.annotations.service.AsyncProcessException;
import com.inteligr8.alfresco.annotations.service.AsyncService;

import jakarta.annotation.PostConstruct;
import jakarta.annotation.PreDestroy;

/**
 * @author brian@inteligr8.com
 */
public abstract class AbstractMqAsyncService extends AbstractLifecycleBean implements AsyncService, InitializingBean, DisposableBean, Threadable {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final JobKey jobKey = new JobKey("mq-async", "inteligr8-annotations");
	protected final Pattern typePattern = Pattern.compile("v([0-9]+):([^:#]+)#(.+)");
	protected final ObjectMapper om = new ObjectMapper();
    
	@Value("${inteligr8.async.mq.enabled}")
	protected boolean enabled;
	
	@Value("${inteligr8.async.mq.workerThreads}")
	protected int workerThreads;
	
    @Value("${inteligr8.async.mq.url}")
    protected String url;
    
    @Value("${inteligr8.async.mq.username}")
    protected String username;
    
    @Value("${inteligr8.async.mq.password}")
    protected String password;
    
    @Value("${inteligr8.async.mq.queue}")
    protected String queueName;
    
    @Value("${inteligr8.async.mq.errorQueue}")
    protected String errorQueueName;
    
    @Value("${inteligr8.async.mq.clientId}")
    protected String clientId;
    
    @Value("${inteligr8.async.mq.pool.max}")
    protected short maxConnections;
    
    @Autowired
    protected ActionService actionService;
    
    @Autowired
    protected ContentService contentService;
    
    @Autowired
    protected CustomModelService modelService;
    
    @Autowired
    protected DictionaryService dictionaryService;
    
    @Autowired
    protected NamespaceService namespaceService;
    
    @Autowired
    protected TransactionService txService;
    
    protected String hostname;
    
    protected SimpleCache<Pair<Class<?>, String>, Method> methodCache;
    
    protected ThreadLocal<Boolean> isAsync = ThreadLocal.withInitial(new Supplier<Boolean>() {
    	@Override
    	public Boolean get() {
    		return false;
    	}
	});
    
    @Override
    public final void afterPropertiesSet() throws Exception {
    	this.init();
    }
    
    @Override
    public final void destroy() throws Exception {
    	this.uninit();
    }
    
    /**
     * @PostConstruct does not work in ACS
     */
    @PostConstruct
    protected void init() {
    	if (!this.enabled)
    		return;
    	
    	try {
    		this.hostname = InetAddress.getLocalHost().getHostName();
    	} catch (UnknownHostException uhe) {
    		this.hostname = "unknown";
    	}

		if (this.workerThreads <= 0)
			throw new AlfrescoRuntimeException("The 'inteligr8.async.mq.workerThreads' property must be positive");
    }
    
    @PreDestroy
    protected void uninit() {
    }
    
    @Override
    protected void onBootstrap(ApplicationEvent event) {
    	if (!this.enabled)
    		return;
    	
    	JobDetailImpl jobDetail = new JobDetailImpl();
    	jobDetail.setKey(this.jobKey);
    	jobDetail.setRequestsRecovery(true);
    	jobDetail.setJobClass(AsyncJob.class);
    	jobDetail.getJobDataMap().put("asyncService", this);
    	
    	Trigger trigger = TriggerBuilder.newTrigger()
    			.startNow()
    			.build();
    	
    	try {
	    	StdSchedulerFactory.getDefaultScheduler()
					.scheduleJob(jobDetail, trigger);
    	} catch (SchedulerException se) {
    		this.logger.error("The MQ async service job failed to start; no asynchronous executions will be processed!", se);
    	}
    }
    
    @Override
    protected void onShutdown(ApplicationEvent event) {
    	try {
	    	Scheduler scheduler = StdSchedulerFactory.getDefaultScheduler();
	    	scheduler.deleteJob(this.jobKey);
    	} catch (SchedulerException se) {
    		this.logger.warn("The MQ async service job failed to stop", se);
    	}
    }
    
    @Override
    public boolean isEnabled() {
		return this.enabled;
	}
    
    @Override
	public Integer getThreads() {
		return this.workerThreads;
	}
    
    @Override
	public boolean isCurrentThreadAsynchronous() {
		return this.isAsync.get();
	}
    
    protected Method findMethod(Class<?> clazz, String methodName) {
    	Pair<Class<?>, String> key = new Pair<>(clazz, methodName);
    	Method method = this.methodCache.get(key);
    	if (method != null) {
			this.logger.trace("Found method in cache: {}", method);
    		return method;
    	}

		this.logger.trace("Looping through bean type methods to find: {}", methodName);
		
		for (Method amethod : clazz.getDeclaredMethods()) {
			if (amethod.getName().equals(methodName)) {
				this.logger.debug("Found and caching method: {} => {}", key, amethod);
				this.methodCache.put(key, amethod);
				return amethod;
			}
		}
		
		throw new IllegalStateException("The bean (" + clazz + ") does not implement the method: " + methodName);
    }

    public void push(ProceedingJoinPoint joinPoint) throws AsyncProcessException {
    	this.logger.trace("push({})", joinPoint);
    	
		if (!(joinPoint.getSignature() instanceof MethodSignature))
			throw new IllegalStateException("The join point must be on methods and methods have signatures");
		
		Object bean = joinPoint.getThis();
		this.logger.debug("Queuing for bean: {}", bean.getClass());

		MethodSignature methodSig = (MethodSignature) joinPoint.getSignature();
		Method method = methodSig.getMethod();
		this.logger.debug("Queuing for method: {}", method);
		
		this.push(bean, method.getName(), Arrays.asList(joinPoint.getArgs()));
    }

	public abstract void push(Object callbackBean, String callbackMethod, List<Object> args) throws AsyncProcessException;
    
    @SuppressWarnings({ "unchecked" })
	protected Object unmarshal(Parameter param, Object arg) throws NoSuchMethodException, IllegalAccessException, InvocationTargetException, InstantiationException {
    	Class<?> paramType = param.getType();
		this.logger.trace("Unmarshaling parameter of type: {}", paramType);

    	if (arg instanceof String || arg instanceof Number || arg instanceof Boolean) {
			this.logger.trace("Unmarshaling primitive: {}", arg);
			return arg;
    	} else if (Temporal.class.isAssignableFrom(paramType)) {
    		if (OffsetDateTime.class.isAssignableFrom(paramType)) {
    			return OffsetDateTime.from(DateTimeFormatter.ISO_OFFSET_DATE_TIME.parse(arg.toString()));
    		} else if (ZonedDateTime.class.isAssignableFrom(paramType)) {
    			return ZonedDateTime.from(DateTimeFormatter.ISO_ZONED_DATE_TIME.parse(arg.toString()));
    		} else if (LocalDate.class.isAssignableFrom(paramType)) {
    			return LocalDate.from(DateTimeFormatter.ISO_LOCAL_DATE.parse(arg.toString()));
    		} else if (LocalDateTime.class.isAssignableFrom(paramType)) {
    			return LocalDateTime.from(DateTimeFormatter.ISO_LOCAL_DATE_TIME.parse(arg.toString()));
    		} else if (Instant.class.isAssignableFrom(paramType)) {
    			return Instant.from(DateTimeFormatter.ISO_INSTANT.parse(arg.toString()));
    		} else if (LocalTime.class.isAssignableFrom(paramType)) {
    			return LocalTime.from(DateTimeFormatter.ISO_LOCAL_TIME.parse(arg.toString()));
    		} else if (OffsetTime.class.isAssignableFrom(paramType)) {
    			return OffsetTime.from(DateTimeFormatter.ISO_OFFSET_TIME.parse(arg.toString()));
    		} else {
    			throw new UnsupportedOperationException();
    		}
    	} else if (Version.class.isAssignableFrom(paramType)) {
			this.logger.trace("Unmarshaling as JSON object: {}", arg);
			Map<String, Object> argMap = (Map<String, Object>) this.om.convertValue(arg, Map.class);
			
			Map<String, Serializable> versionPropertiesMap = (Map<String, Serializable>) argMap.get("properties");
			NodeRef nodeRef = new NodeRef((String) argMap.get("nodeRef"));
			
			Version version = new VersionImpl(versionPropertiesMap, nodeRef);
			this.logger.trace("Unmarshaled version: {} = {}", param.getName(), version);
			return version;
		} else if (Action.class.isAssignableFrom(paramType)) {
			this.logger.trace("Unmarshaling as JSON object: {}", arg);
			Map<String, Object> argMap = (Map<String, Object>) this.om.convertValue(arg, Map.class);
			
			String actionId = (String) argMap.get("actionId");
			NodeRef nodeRef = new NodeRef((String) argMap.get("nodeRef"));
			this.logger.trace("Unmarshaling action: {}, {}", actionId, nodeRef);
			
			Action action = this.actionService.getAction(nodeRef, actionId);
			this.logger.trace("Unmarshaled action: {} = {}", param.getName(), action);
			return action;
		} else if (Collection.class.isAssignableFrom(paramType)) {
			this.logger.trace("Unmarshaling as JSON array: {}", arg);
			return this.om.convertValue(arg, Collection.class);
		} else if (Map.class.isAssignableFrom(paramType)) {
			this.logger.trace("Unmarshaling as JSON object: {}", arg);
			return this.om.convertValue(arg, Map.class);
		} else if (QName.class.isAssignableFrom(paramType)) {
			this.logger.trace("Unmarshaling as QName: {}", arg);
			return QName.createQName((String) arg);
		} else if (Enum.class.isAssignableFrom(paramType)) {
			this.logger.trace("Unmarshaling as Enum: {}", arg);
			Method cons = paramType.getDeclaredMethod("valueOf", String.class);
			return cons.invoke(null, arg.toString());
		} else {
			this.logger.trace("Unmarshaling as POJO: {}", arg);
			try {
				Constructor<?> cons = paramType.getConstructor(String.class);
				return cons.newInstance(arg.toString());
			} catch (NoSuchMethodException nsme) {
				Method method = paramType.getDeclaredMethod("valueOf", String.class);
				return method.invoke(null, arg.toString());
			}
		}
    }
    
    protected Object marshal(Object arg) {
    	if (arg instanceof String || arg instanceof Number || arg instanceof Boolean) {
    		return arg;
    	} else if (arg instanceof Temporal) {
    		return arg.toString();
    	} else if (arg instanceof Version) {
			Version version = (Version) arg;
			Map<String, Object> map = new HashMap<>();
			map.put("nodeRef", version.getFrozenStateNodeRef());
			map.put("properties", version.getVersionProperties());

			this.logger.trace("Marshaling Version as JSON object: {}", map);
			return this.om.convertValue(map, String.class);
		} else if (arg instanceof Action) {
			Action action = (Action) arg;
			Map<String, Object> map = new HashMap<>();
			map.put("nodeRef", action.getNodeRef());
			map.put("actionId", action.getId());

			this.logger.trace("Marshaling Action as JSON object: {}", map);
			return this.om.convertValue(map, String.class);
		} else if (arg instanceof Collection<?>) {
			List<Object> list = new ArrayList<>(((Collection<?>)arg).size());
			for (Object obj : (Collection<?>) arg)
				list.add(this.marshal(obj));
			
			this.logger.trace("Marshaling Java Collection as JSON array: {}", list);
			return this.om.convertValue(list, String.class);
		} else if (arg instanceof Map<?, ?>) {
			Map<Object, Object> map = new HashMap<>();
			for (Entry<?, ?> entry : ((Map<?, ?>) arg).entrySet()) {
				Object key = this.marshal(entry.getKey());
				Object value = this.marshal(entry.getValue());
				map.put(key, value);
			}

			this.logger.trace("Marshaling Java Map as JSON object: {}", map);
			return this.om.convertValue(map, String.class);
		} else {
			this.logger.trace("Marshaling Java object as JSON object: {}", arg);
			return this.om.convertValue(arg, String.class);
		}
    }
    
    protected M2Model loadModel(NodeRef nodeRef) throws IOException {
		ContentReader creader = this.contentService.getReader(nodeRef, ContentModel.PROP_CONTENT);
		InputStream istream = creader.getContentInputStream();
		try {
			return M2Model.createModel(istream);
		} finally {
			istream.close();
		}
    }

}
