package com.inteligr8.alfresco.annotations.service.impl;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;

import org.alfresco.repo.security.authentication.AuthenticationUtil;
import org.alfresco.repo.security.authentication.AuthenticationUtil.RunAsWork;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.reflect.MethodSignature;
import org.quartz.JobKey;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.Trigger;
import org.quartz.TriggerBuilder;
import org.quartz.impl.JobDetailImpl;
import org.quartz.impl.StdSchedulerFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEvent;
import org.springframework.extensions.surf.util.AbstractLifecycleBean;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import com.inteligr8.alfresco.annotations.job.AsyncJob;
import com.inteligr8.alfresco.annotations.service.AsyncProcessException;
import com.inteligr8.alfresco.annotations.service.AsyncService;

import jakarta.annotation.PostConstruct;

/**
 * This class provides a non-persistent alternative to MQ for asynchronous method
 * execution.
 * 
 * @author brian@inteligr8.com
 */
@Component("async.thread")
public class ThreadPoolAsyncService extends AbstractLifecycleBean implements AsyncService, InitializingBean {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final JobKey jobKey = new JobKey("thread-async", "inteligr8-annotations");
	
    @Value("${inteligr8.async.workableThreads}")
    protected int queueSize;
	
    @Value("${inteligr8.async.workerThreads}")
    protected byte poolSize;
    
    private LinkedBlockingQueue<Runnable> queue;
    private ThreadPoolExecutor pool;
    
    private ThreadLocal<Boolean> isAsync = ThreadLocal.withInitial(new Supplier<Boolean>() {
    	@Override
    	public Boolean get() {
    		return false;
    	}
	});
    
    @Override
    public void afterPropertiesSet() throws Exception {
    	this.init();
	}
	
	/**
	 * @PostConstruct doesn't work in ACS for whatever reason
	 */
	@PostConstruct
	protected void init() {
    	this.queue = new LinkedBlockingQueue<>(this.queueSize);
    }
    
    protected void onBootstrap(ApplicationEvent event) {
    	JobDetailImpl jobDetail = new JobDetailImpl();
    	jobDetail.setKey(this.jobKey);
    	jobDetail.setRequestsRecovery(true);
    	jobDetail.setJobClass(AsyncJob.class);
    	jobDetail.getJobDataMap().put("asyncService", this);
    	
    	Trigger trigger = TriggerBuilder.newTrigger()
    			.startNow()
    			.build();
    	
    	try {
	    	StdSchedulerFactory.getDefaultScheduler()
	    			.scheduleJob(jobDetail, trigger);
    	} catch (SchedulerException se) {
    		this.logger.error("The async service job failed to start; no asynchronous executions will be processed!", se);
    	}
    }
    
    @Override
    protected void onShutdown(ApplicationEvent event) {
    	try {
	    	Scheduler scheduler = StdSchedulerFactory.getDefaultScheduler();
	    	scheduler.deleteJob(this.jobKey);
    	} catch (SchedulerException se) {
    		this.logger.warn("The MQ async service job failed to stop", se);
    	}
    }
    
    @Override
    public boolean isEnabled() {
    	return true;
	}
    
    @Override
	public boolean isCurrentThreadAsynchronous() {
		return this.isAsync.get();
	}
    
    @Override
    public void poll() throws AsyncProcessException {
    	this.logger.trace("poll()");
		this.isAsync.set(true);

    	this.pool = new ThreadPoolExecutor(1, this.poolSize, 5L, TimeUnit.SECONDS, this.queue);

		try {
			while (!this.pool.isTerminated()) {
				this.logger.debug("Perpetually waiting for thread pool to terminate ...");
    			this.pool.awaitTermination(300L, TimeUnit.SECONDS);
    		}

			this.logger.info("The Async service thread pool terminated");
		} catch (InterruptedException ie) {
			this.logger.info("The Async service thread pool was interrupted");
    	}
    }
    
    @Transactional
    protected void executeWork(Object callbackBean, Method callbackMethod, List<Object> args) {
    	try {
    		callbackMethod.invoke(callbackBean, args.toArray());
		} catch (IllegalAccessException iae) {
			this.logger.error("A bean method was not accessible (public)");
			this.logger.warn("The bean '{}' method '{}' is not accessible: {}", callbackBean.getClass(), callbackMethod, iae.getMessage());
		} catch (InvocationTargetException ite) {
			this.logger.error("A bean method execution failed");
			this.logger.warn("The bean '{}' method '{}' execution failed: {}", callbackBean.getClass(), callbackMethod, ite.getMessage());
    	}
    }

    public void push(ProceedingJoinPoint joinPoint) throws AsyncProcessException {
    	this.logger.trace("push({})", joinPoint);
    	
		if (!(joinPoint.getSignature() instanceof MethodSignature))
			throw new IllegalStateException("The join point must be on methods and methods have signatures");
		
		Object bean = joinPoint.getThis();
		this.logger.debug("Queuing for bean: {}", bean.getClass());

		MethodSignature methodSig = (MethodSignature) joinPoint.getSignature();
		Method method = methodSig.getMethod();
		this.logger.debug("Queuing for method: {}", method);
		
		this.push(bean, method, Arrays.asList(joinPoint.getArgs()));
    }
    
    public void push(Object callbackBean, Method callbackMethod, List<Object> args) throws AsyncProcessException {
    	this.logger.trace("push({}, {}, {})", callbackBean, callbackMethod, args);
    	
    	RunAsWork<Void> work = new RunAsWork<Void>() {
    		@Override
    		public Void doWork() {
				executeWork(callbackBean, callbackMethod, args);
				return null;
    		}
		};

    	final String runAs = AuthenticationUtil.getRunAsUser();
    	Runnable runnable = new Runnable() {
			@Override
			public void run() {
				// run the thread with the same authentication context as the initiating user
				AuthenticationUtil.runAs(work, runAs);
			}
		};
		
    	this.queue.offer(runnable);
    }

}
