package com.inteligr8.alfresco.annotations;

import org.alfresco.repo.transaction.AlfrescoTransactionSupport;
import org.alfresco.repo.transaction.AlfrescoTransactionSupport.TxnReadState;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationEvent;
import org.springframework.extensions.surf.util.AbstractLifecycleBean;
import org.springframework.transaction.IllegalTransactionStateException;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

public class AbstractTransactionalTest extends AbstractLifecycleBean {
	
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Override
	protected void onBootstrap(ApplicationEvent event) {
		this.logger.info("Running test: " + this.getClass());
		Assert.isNull(AlfrescoTransactionSupport.getTransactionId(), "An unexpected transaction: " + AlfrescoTransactionSupport.getTransactionId());

		this.tryOutsideTx();
		this.tryWithinTx();
		this.tryWithinReadonlyTx();
	}
	
	@Override
	protected void onShutdown(ApplicationEvent event) {
	}
	
	private void tryOutsideTx() {
		this.logger.info("Running outside TX test");
		
		this.tryDefaultTransactional(null, false);
		this.tryReadOnlyTransactional(null, false);
		this.tryRetryOnlyTransactional(null);
		this.trySupportsTransactional(null, false);
		this.tryRequiresNewTransactional(null, false);
		this.tryRequiredTransactional(null, false);
		this.tryNeverTransactional(null);
		
		try {
			this.tryNoSupportsTransactional();
		} catch (IllegalTransactionStateException uoe) {
			throw new IllegalStateException();
		}
		
		try {
			this.tryMandatoryTransactional(null, false);
			throw new IllegalStateException();
		} catch (IllegalTransactionStateException itse) {
			// suppress
		}
	}
	
	@Transactional
	private void tryWithinTx() {
		this.logger.info("Running inside read/write TX test");
		
		String txId = AlfrescoTransactionSupport.getTransactionId();
		boolean readonly = false;

		this.tryDefaultTransactional(txId, readonly);
		this.tryReadOnlyTransactional(txId, readonly);
		this.tryRetryOnlyTransactional(txId);
		this.trySupportsTransactional(txId, readonly);
		this.tryRequiresNewTransactional(txId, readonly);
		this.tryRequiredTransactional(txId, readonly);
		this.tryMandatoryTransactional(txId, readonly);
		
		try {
			this.tryNoSupportsTransactional();
			throw new IllegalStateException();
		} catch (IllegalTransactionStateException uoe) {
			// suppress
		}
		
		try {
			this.tryNeverTransactional(txId);
			throw new IllegalStateException();
		} catch (IllegalTransactionStateException itse) {
			// suppress
		}
	}
	
	@Transactional(readOnly = true)
	private void tryWithinReadonlyTx() {
		this.logger.info("Running inside read-only TX test");
		
		String txId = AlfrescoTransactionSupport.getTransactionId();
		boolean readonly = true;

		this.tryDefaultTransactional(txId, readonly);
		this.tryReadOnlyTransactional(txId, readonly);
		this.tryRetryOnlyTransactional(txId);
		this.trySupportsTransactional(txId, readonly);
		this.tryRequiresNewTransactional(txId, readonly);
		this.tryRequiredTransactional(txId, readonly);
		
		try {
			this.tryNoSupportsTransactional();
			throw new IllegalStateException();
		} catch (IllegalTransactionStateException uoe) {
			// suppress
		}
		
		try {
			this.tryMandatoryTransactional(txId, readonly);
			throw new IllegalStateException();
		} catch (IllegalTransactionStateException itse) {
			// suppress
		}
		
		try {
			this.tryNeverTransactional(txId);
			throw new IllegalStateException();
		} catch (IllegalTransactionStateException itse) {
			// suppress
		}
	}
	
	@Transactional
	private void tryDefaultTransactional(String originTxId, boolean originReadonly) {
		Assert.hasText(AlfrescoTransactionSupport.getTransactionId(), "Expected a transaction");
		Assert.isTrue(TxnReadState.TXN_READ_WRITE.equals(AlfrescoTransactionSupport.getTransactionReadState()), "Expected a read/write transaction");
		if (originTxId != null) {
			if (originReadonly) {
				// changed from readonly to read/write; need new TX
				Assert.isTrue(!AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected a different transaction: " + AlfrescoTransactionSupport.getTransactionId() + " == " + originTxId);
			} else {
				// no changes; same TX
				Assert.isTrue(AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected the same transaction: " + AlfrescoTransactionSupport.getTransactionId() + " != " + originTxId);
			}
		}
	}
	
	@Transactional(readOnly = true)
	private void tryReadOnlyTransactional(String originTxId, boolean originReadonly) {
		Assert.hasText(AlfrescoTransactionSupport.getTransactionId(), "Expected a transaction");
		Assert.isTrue(TxnReadState.TXN_READ_ONLY.equals(AlfrescoTransactionSupport.getTransactionReadState()), "Expected a readonly transaction");
		if (originTxId != null) {
			if (originReadonly) {
				// no changes; same TX
				Assert.isTrue(AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected the same transaction: " + AlfrescoTransactionSupport.getTransactionId() + " != " + originTxId);
			} else {
				// changed from read/write to readonly; need new TX
				Assert.isTrue(!AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected a different transaction: " + AlfrescoTransactionSupport.getTransactionId() + " == " + originTxId);
			}
		}
	}
	
	@Transactional(propagation = Propagation.SUPPORTS)
	private void trySupportsTransactional(String originTxId, boolean originReadonly) {
		if (originTxId == null) {
			Assert.isNull(AlfrescoTransactionSupport.getTransactionId(), "Unexpected transaction");
		} else {
			Assert.hasText(AlfrescoTransactionSupport.getTransactionId(), "Expected a transaction");
			Assert.isTrue(originReadonly == TxnReadState.TXN_READ_ONLY.equals(AlfrescoTransactionSupport.getTransactionReadState()), "Expected the same read-state transaction");
			Assert.isTrue(AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected the same transaction: " + AlfrescoTransactionSupport.getTransactionId() + " != " + originTxId);
		}
	}
	
	@Transactional(propagation = Propagation.REQUIRED)
	private void tryRequiredTransactional(String originTxId, boolean originReadonly) {
		Assert.hasText(AlfrescoTransactionSupport.getTransactionId(), "Expected a transaction");
		Assert.isTrue(TxnReadState.TXN_READ_WRITE.equals(AlfrescoTransactionSupport.getTransactionReadState()), "Expected a read/write transaction");
		if (originTxId != null) {
			if (originReadonly) {
				// changed from readonly to read/write; need new TX
				Assert.isTrue(!AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected a different transaction: " + AlfrescoTransactionSupport.getTransactionId() + " == " + originTxId);
			} else {
				// no changes; same TX
				Assert.isTrue(AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected the same transaction: " + AlfrescoTransactionSupport.getTransactionId() + " != " + originTxId);
			}
		}
	}
	
	@Transactional(propagation = Propagation.REQUIRES_NEW)
	private void tryRequiresNewTransactional(String originTxId, boolean originReadonly) {
		Assert.hasText(AlfrescoTransactionSupport.getTransactionId(), "Expected a transaction");
		Assert.isTrue(TxnReadState.TXN_READ_WRITE.equals(AlfrescoTransactionSupport.getTransactionReadState()), "Expected a read/write transaction");
		if (originTxId != null)
			Assert.isTrue(!AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected a different transaction: " + AlfrescoTransactionSupport.getTransactionId() + " == " + originTxId);
	}
	
	@Transactional(propagation = Propagation.MANDATORY)
	private void tryMandatoryTransactional(String originTxId, boolean originReadonly) {
		if (originTxId == null) {
			throw new IllegalStateException();
		} else {
			Assert.hasText(AlfrescoTransactionSupport.getTransactionId(), "Expected a transaction");
			Assert.isTrue(AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected the same transaction: " + AlfrescoTransactionSupport.getTransactionId() + " != " + originTxId);
			Assert.isTrue(originReadonly == TxnReadState.TXN_READ_ONLY.equals(AlfrescoTransactionSupport.getTransactionReadState()), "Expected the same read-state transaction");
		}
	}
	
	@Transactional(propagation = Propagation.NOT_SUPPORTED)
	private void tryNoSupportsTransactional() {
		Assert.isNull(AlfrescoTransactionSupport.getTransactionId(), "Expected no transaction");
		Assert.isTrue(TxnReadState.TXN_NONE.equals(AlfrescoTransactionSupport.getTransactionReadState()), "Expected not transaction");
	}
	
	@Transactional(propagation = Propagation.NEVER)
	private void tryNeverTransactional(String originTxId) {
		if (originTxId == null) {
			Assert.isNull(AlfrescoTransactionSupport.getTransactionId(), "Unexpected transaction");
		} else {
			throw new IllegalStateException();
		}
	}
	
	@TransactionalRetryable
	private void tryRetryOnlyTransactional(String originTxId) {
		if (originTxId == null) {
			Assert.isTrue(AlfrescoTransactionSupport.getTransactionId() != null, "Expected a new transaction");
		} else {
			Assert.isTrue(AlfrescoTransactionSupport.getTransactionId().equals(originTxId), "Expected the same transaction: " + AlfrescoTransactionSupport.getTransactionId() + " != " + originTxId);
		}
	}

}
