package com.inteligr8.alfresco.annotations.service.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Parameter;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.OffsetDateTime;
import java.time.OffsetTime;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.Temporal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.function.Supplier;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.alfresco.error.AlfrescoRuntimeException;
import org.alfresco.model.ContentModel;
import org.alfresco.repo.cache.SimpleCache;
import org.alfresco.repo.dictionary.M2Model;
import org.alfresco.repo.version.common.VersionImpl;
import org.alfresco.service.cmr.action.Action;
import org.alfresco.service.cmr.action.ActionService;
import org.alfresco.service.cmr.dictionary.CustomModelService;
import org.alfresco.service.cmr.dictionary.DictionaryService;
import org.alfresco.service.cmr.repository.ContentReader;
import org.alfresco.service.cmr.repository.ContentService;
import org.alfresco.service.cmr.repository.NodeRef;
import org.alfresco.service.cmr.version.Version;
import org.alfresco.service.namespace.NamespaceService;
import org.alfresco.service.namespace.QName;
import org.alfresco.service.transaction.TransactionService;
import org.alfresco.util.Pair;
import org.apache.activemq.ActiveMQConnectionFactory;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.reflect.MethodSignature;
import org.messaginghub.pooled.jms.JmsPoolConnectionFactory;
import org.quartz.JobKey;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.Trigger;
import org.quartz.TriggerBuilder;
import org.quartz.impl.JobDetailImpl;
import org.quartz.impl.StdSchedulerFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEvent;
import org.springframework.extensions.surf.util.AbstractLifecycleBean;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.inteligr8.alfresco.annotations.AuthorizedAsSystem;
import com.inteligr8.alfresco.annotations.Threadable;
import com.inteligr8.alfresco.annotations.Threaded;
import com.inteligr8.alfresco.annotations.TransactionalRetryable;
import com.inteligr8.alfresco.annotations.job.AsyncJob;
import com.inteligr8.alfresco.annotations.service.AsyncProcessException;
import com.inteligr8.alfresco.annotations.service.AsyncService;

import jakarta.annotation.PostConstruct;
import jakarta.annotation.PreDestroy;
import jakarta.jms.Connection;
import jakarta.jms.JMSException;
import jakarta.jms.Message;
import jakarta.jms.MessageConsumer;
import jakarta.jms.MessageProducer;
import jakarta.jms.Queue;
import jakarta.jms.Session;

/**
 * This class provides integration with MQ for the asynchronous method executions.
 * 
 * @author brian@inteligr8.com
 */
@Component("async.mq")
public class MqAsyncService extends AbstractMqAsyncService {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());

    private JmsPoolConnectionFactory factory;
    
    /**
     * @PostConstruct does not work in ACS
     */
    @PostConstruct
    protected void init() {
    	if (!this.enabled)
    		return;

		super.init();
    	
    	ActiveMQConnectionFactory factory = new ActiveMQConnectionFactory(this.url);
    	
    	JmsPoolConnectionFactory pool = new JmsPoolConnectionFactory();
    	pool.setConnectionFactory(factory);
    	pool.setMaxConnections(this.maxConnections);
    	pool.start();
    	
    	this.factory = pool;
    }
    
    @PreDestroy
    protected void uninit() {
		super.uninit();

    	if (this.factory != null)
    		this.factory.stop();
    }
    
    @Override
    @Transactional
    public void poll() throws AsyncProcessException {
    	this.logger.trace("poll()");
		this.isAsync.set(true);
		
    	try {
	    	Connection mqcon = this.factory.createConnection(this.username, this.password);
	    	try {
		    	mqcon.setClientID(this.clientId + "-service-" + this.hostname);
	
		    	Session mqsession = mqcon.createSession(true, Session.CLIENT_ACKNOWLEDGE);
		    	try {
			    	this.logger.debug("Polling messages for asynchronous policy execution");
			    	this.pollErrors(mqsession);
			    	this.pollMain(mqsession);
		    	} finally {
		    		mqsession.close();
		    	}
	    	} finally {
	    		mqcon.close();
	    	}
    	} catch (JMSException je) {
    		throw new AsyncProcessException("A JMS messaging issue occurred", je);
    	}
    }
    
    private void pollErrors(Session mqsession) throws JMSException {
    	this.logger.debug("Polling previously errored messages");
    	
    	Queue mqqueue = mqsession.createQueue(this.errorQueueName);
    	Set<String> msgIds = new HashSet<>();
    	int ackd = 0;
    	
		MessageConsumer consumer = mqsession.createConsumer(mqqueue);
		try {
			while (!Thread.currentThread().isInterrupted()) {
				Boolean processed = this.pollTx(mqsession, consumer, msgIds);
				if (processed == null) {
					break;
				} else if (processed.booleanValue()) {
					ackd++;
				}
			}
		} finally {
			consumer.close();
		}

		this.logger.info("Successfully processed {} of {} previously errored messages", ackd, msgIds.size());
    }
    
    private void pollMain(Session mqsession) throws JMSException {
    	this.logger.debug("Polling ongoing messages ...");
    	
    	Queue mqqueue = mqsession.createQueue(this.queueName);
    	this.pollMainThreaded(mqsession, mqqueue);
    }
    
    @Threaded(name = "mq-poll", join = true)
    private void pollMainThreaded(Session mqsession, Queue mqqueue) throws JMSException {
		MessageConsumer consumer = mqsession.createConsumer(mqqueue);
		try {
			while (!Thread.currentThread().isInterrupted()) {
				pollTx(mqsession, consumer, null);
			}
		} finally {
			consumer.close();
		}
    }

    @Transactional(propagation = Propagation.REQUIRES_NEW)
    @TransactionalRetryable(maxRetries = 3)
    @AuthorizedAsSystem
    private Boolean pollTx(Session mqsession, MessageConsumer consumer, Set<String> msgIds) throws JMSException {
		Message mqmsg = consumer.receive();
		
		if (msgIds != null && !msgIds.add(mqmsg.getJMSMessageID())) {
			this.logger.debug("Received a message again; assuming we have (re)tried all errored messages: {}", mqmsg.getJMSMessageID());
			return null;
		}
		
		try {
			if (this.processIncomingMessage(mqsession, mqmsg, msgIds != null)) {
				mqmsg.acknowledge();
				return true;
			}
		} catch (RuntimeException | Error e) {
			this.logger.error("An unexpected issue occurred", e);
		}
		
		 return false;
    }
    
    private boolean processIncomingMessage(Session mqsession, Message mqmsg, boolean isErrorQueue) throws JMSException {
		String msgId = mqmsg.getJMSMessageID();
		this.logger.debug("Received message: {}", msgId);
		
		String type = mqmsg.getJMSType();
		Matcher matcher = this.typePattern.matcher(type);
		if (!matcher.find()) {
			this.logger.warn("The queue has a message ('{}') with an unsupported JMS type: {}", msgId, type);
			return false;
		}
		
		try {
			Class<?> beanClass = Class.forName(matcher.group(2));
			this.logger.trace("Preparing to execute using bean type: {}", beanClass);
			Object bean = this.getApplicationContext().getBean(beanClass);
			this.logger.trace("Found qualifying bean: {}", bean);
			
			String methodName = matcher.group(3);
			Method method = this.findMethod(beanClass, methodName);
			this.logger.trace("Found qualifying method: {}", method);
			Parameter[] params = method.getParameters();
			
			Object[] args = new Object[params.length];
			
			for (int a = 0; a < args.length; a++) {
				Object arg = mqmsg.getObjectProperty("arg" + a);
				if (arg == null)
					continue;
				
				args[a] = this.unmarshal(params[a], arg);
			}
			
			switch (method.getName()) {
				case "onLoadDynamicModel":
					args[1] = args[0];
					args[0] = this.loadModel((NodeRef) args[1]);
			}
			
			method.invoke(bean, args);
		} catch (ClassNotFoundException cnfe) {
			this.logger.error("A bean could not be found; will try on next restart");
			this.logger.error("The bean '{}' could not be found: {}", matcher.group(2), cnfe.getMessage());
			if (isErrorQueue)
				return false;
			this.moveToErrorQueue(mqsession, mqmsg);
		} catch (IOException ie) {
			this.logger.warn("This should never happen: " + ie.getMessage());
			// return to queue and retry indefinitely
			return false;
		} catch (NoSuchMethodException nsme) {
			this.logger.error("A bean enum argument could not be constructed; will try on next restart");
			this.logger.error("An argument could not be The bean '{}' could not be found: {}", matcher.group(2), nsme.getMessage());
			if (isErrorQueue)
				return false;
			this.moveToErrorQueue(mqsession, mqmsg);
		} catch (IllegalAccessException iae) {
			this.logger.error("A bean method was not accessible (public); will try on next restart");
			this.logger.warn("The bean '{}' method '{}' is not accessible: {}", matcher.group(2), matcher.group(3), iae.getMessage());
			if (isErrorQueue)
				return false;
			this.moveToErrorQueue(mqsession, mqmsg);
		} catch (InstantiationException | InvocationTargetException ie) {
			this.logger.error("A bean method execution failed; will try on next restart");
			this.logger.warn("The bean '{}' method '{}' execution failed: {}", matcher.group(2), matcher.group(3), ie.getMessage());
			if (isErrorQueue)
				return false;
			this.moveToErrorQueue(mqsession, mqmsg);
		}

		return true;
    }
    
    private void moveToErrorQueue(Session mqsession, Message mqmsg) throws JMSException {
    	Queue mqqueue = mqsession.createQueue(this.errorQueueName);
    	
    	MessageProducer producer = mqsession.createProducer(mqqueue);
    	try {
    		producer.send(mqmsg);
    	} finally {
    		producer.close();
    	}
    }
    
    @Transactional
    public void push(Object callbackBean, String callbackMethod, List<Object> args) throws AsyncProcessException {
    	this.logger.trace("push({}, {}, {})", callbackBean.getClass(), callbackMethod, args);
    	
    	UUID msgId = UUID.randomUUID();
    	
    	try {
	    	Connection mqcon = this.factory.createConnection(this.username, this.password);
	    	try {
		    	mqcon.setClientID(this.clientId + "-client-" + this.hostname);
		    	
		    	Session mqsession = mqcon.createSession(true, Session.AUTO_ACKNOWLEDGE);
		    	try {
			    	this.logger.trace("Sending policy as message: {} => {}", callbackMethod, msgId);
	
			    	Queue mqqueue = mqsession.createQueue(this.queueName);
			    	
			    	Message mqmsg = mqsession.createMessage();
			    	mqmsg.setJMSMessageID(msgId.toString());
			    	mqmsg.setJMSType("v1:" + callbackBean.getClass() + "#" + callbackMethod);
			    	
			    	int i = 0;
			    	for (Object arg : args)
			    		mqmsg.setObjectProperty("arg" + (i++), this.marshal(arg));
			    	
			    	MessageProducer producer = mqsession.createProducer(mqqueue);
			    	try {
			    		producer.send(mqmsg);
			    	} finally {
			    		producer.close();
			    	}
	
			    	this.logger.debug("Sent node as message: {} => {}", callbackMethod, msgId);
		    	} finally {
			    	mqsession.close();
		    	}
	    	} finally {
	    		mqcon.close();
	    	}
    	} catch (JMSException je) {
    		throw new AsyncProcessException("A JMS messaging issue occurred", je);
    	}
    }

}
